//
//  MGenericCanvasView.h
//  GenericCanvas
//
//  Created by Alfredo Kojima on 05/6/16.
//  Copyright 2005 MySQL AB. All rights reserved.
//

#import <Cocoa/Cocoa.h>

class MGCVListener;
class CGenericCanvas;
class CGCView;

@class MCanvasScrollView;
@class MGenericCanvasView;

@protocol MGenericCanvasProtocol
- (BOOL)canvas:(MGenericCanvasView*)canvas handleMouseUp:(int)button atPoint:(NSPoint)pos;
- (BOOL)canvas:(MGenericCanvasView*)canvas handleMouseDown:(int)button atPoint:(NSPoint)pos;
- (BOOL)canvas:(MGenericCanvasView*)canvas handleMouseDrag:(NSPoint)pos;
@end

@interface MGenericCanvasView : NSOpenGLView 
{
  CGenericCanvas *_canvas;
  MGCVListener *_listener;
  
  NSSize _baseSize;
  NSPoint _offset;
  
  NSPoint _lastClickPoint;
  
  NSCursor *_cursor;
  
  NSRect _viewport;
  NSPoint _zoom;
  int _currentZoomStep;
  int _autoScrollInterval;
  float _autoScrollDelay;
  
  id<MGenericCanvasProtocol> _delegate;
  
  BOOL _autoCenterZoom;
  BOOL _grabPanning;

  struct {
    unsigned panning:1;
  } _state;
}

- (void)setZoom:(NSPoint)zoom withCenter:(NSPoint)center;
- (void)setZoomLevel:(int)level;
- (int)zoomLevel;

- (void)setDelegate:(id)delegate;

- (void)setCursor:(NSCursor*)cursor;

- (void)setOffset:(NSPoint)offset;
- (void)setBaseSize:(NSSize)size;
- (NSSize)baseSize;

- (NSSize)actualSize;
- (NSRect)actualVisibleRect;

- (void)zoomInAt:(NSPoint)center;
- (void)zoomOutAt:(NSPoint)center;
- (void)setZoom:(NSPoint)zoom;

- (BOOL)loadLayoutsFromFile:(NSString*)file;
- (BOOL)loadStylesFromFile:(NSString*)file withVariables:(NSDictionary*)variables;

- (void)renderScene;
- (void)updateScrollers;

- (void)setGrabPanning:(BOOL)flag;

- (void)createCanvasView:(NSString*)viewName;
- (void)createCanvasViewWithView:(CGCView*)view;

- (void)switchToCanvasView:(CGCView*)view;

- (void)startAreaSelectionAt:(NSPoint)point;
- (NSRect)finishAreaSelection;

- (CGenericCanvas*)canvas;

- (NSArray*)getZoomSteps;
@end
