
#import "MGRTTableEditor.h"
#import <MySQLToolsCommon/MAccessoryScrollView.h>
#import <MySQLToolsCommon/NSView_extras.h>
#import <MySQLToolsCommon/mxUtils.h>
#import <MySQLToolsCommon/myxutil.h>
#include <MySQLGRT/MGRTValue.h>

#define COLUMN_TYPE(c) ((c).datatype_pointer ?  (c).datatype_pointer->group : 0)

NSString *MGRTTableEditorTableDidChange= @"MGRTTableEditorTableDidChange";
NSString *MGRTTableRowPboardType= @"MGRTTableRowPboardType";

@interface MGRTTableEditor(Private)
- (void)addColumn:(id)sender;
- (void)deleteColumn:(id)sender;
- (void)showColumn:(MGRTValue*)column;
- (void)addFK:(id)sender;
- (void)deleteFK:(id)sender;
- (void)showFK:(MGRTValue*)fk;
- (void)addFKColumn:(id)sender;
- (void)deleteFKColumn:(id)sender;
- (void)addIndex:(id)sender;
- (void)deleteIndex:(id)sender;
- (void)showIndex:(MGRTValue*)index;
- (void)addIndexColumn:(id)sender;
- (void)deleteIndexColumn:(id)sender;

- (void)setupCollationsForCharset:(const char*)charset
                            popup:(id)popup;
- (void)handleTextEdited:(NSNotification*)notif;

- (void)tableViewSelectionDidChange:(NSNotification *)aNotification;
- (BOOL)commitChanges;
- (void)showObject;
- (void)refreshForTableType;
- (void)columnListChanged;
@end


@implementation MGRTTableEditor(Private)

- (void)addColumn:(id)sender
{
  _tableData->addColumn("new_column", "INT");

  [columnList reloadData];
  [columnList selectRow:_tableData->columnCount()-1 byExtendingSelection:NO];
  [columnList editColumn:0 row:_tableData->columnCount()-1 withEvent:nil select:YES];

  [self columnListChanged];
}

- (void)deleteColumn:(id)sender
{
  int i= [columnList selectedRow];
  if (i>=0)
  {
    _tableData->removeColumn(i);
    
    [columnList reloadData];    
    [self columnListChanged];
  }
}

- (void)columnListChanged
{
  NSPopUpButtonCell *pcell= [[indColumnList tableColumnWithIdentifier:@"column"] dataCell];
  NSPopUpButtonCell *pcell2= [[fkColumnList tableColumnWithIdentifier:@"column"] dataCell];
  int i;
  
  [pcell removeAllItems];
  [pcell2 removeAllItems];
  for (i= 0; i < _tableData->columnCount(); i++)
  {
    const char *name= _tableData->getColumn(i)["name"];
    NSString *str= NSStr(name);
    [pcell addItemWithTitle:str];
    [pcell2 addItemWithTitle:str];
  }
}

- (void)addFK:(id)sender
{
  _tableData->addFK("new_fk");

  [fkList reloadData];
  [fkList selectRow:_tableData->fkCount()-1 byExtendingSelection:NO];  
}

- (void)deleteFK:(id)sender
{
  int i= [fkList selectedRow];
  if (i>=0)
  {
    _tableData->removeFK(i);
    
    [fkList reloadData];
  }
}

- (void)addFKColumn:(id)sender
{
  int index= [fkList selectedRow];
  int columnIndex= [columnList selectedRow];

  if (index < 0)
    return;
  
  MGRTValue fk(_tableData->getFK(index));
  
  if (columnIndex < 0)
    _tableData->addColumnToFK(fk);
  else
    _tableData->addColumnToFK(fk, 
                              _tableData->getColumn(columnIndex),
                              "");

  [fkColumnList reloadData];
  [fkColumnList selectRow:fk["columns"].count()-1 byExtendingSelection:NO];  
  [fkColumnList editColumn:0 row:fk["columns"].count()-1 withEvent:nil select:YES];  
}

- (void)deleteFKColumn:(id)sender
{
  int fkindex= [fkList selectedRow];
  int fcolumn= [fkColumnList selectedRow];

  if (fkindex < 0 || fcolumn < 0)
    return;
  
  MGRTValue fk(_tableData->getFK(fkindex));
  
  _tableData->removeColumnFromFK(fk, fcolumn);
  
  [fkColumnList reloadData];
}

- (void)addIndex:(id)sender
{
  _tableData->addIndex("new_index");
  
  [indexList reloadData];
  [indexList selectRow:_tableData->indexCount()-1 byExtendingSelection:NO];
}

- (void)deleteIndex:(id)sender
{
  int i= [indexList selectedRow];
  if (i>=0)
  {
    _tableData->removeIndex(i);
    
    [indexList reloadData];
  }
}

- (void)addIndexColumn:(id)sender
{
  int indexi= [indexList selectedRow];
  int column= [columnList selectedRow];
  
  if (indexi < 0)
    return;

  MGRTValue index(_tableData->getIndex(indexi));
  
  if (column < 0)
    _tableData->addColumnToIndex(index);
  else
    _tableData->addColumnToIndex(index, _tableData->getColumn(column));
    
  [indColumnList reloadData];
  [indColumnList selectRow:index["columns"].count()-1 byExtendingSelection:NO];  
  [indColumnList editColumn:0 row:index["columns"].count()-1 withEvent:nil select:YES];
}

- (void)deleteIndexColumn:(id)sender
{
  int index= [indexList selectedRow];
  int icolumn= [indColumnList selectedRow];

  if (index < 0 || icolumn < 0)
    return;
  
  _tableData->removeColumnFromIndex(_tableData->getIndex(index), icolumn);

  [indColumnList reloadData];
}

- (void)handleTextEdited:(NSNotification*)notif
{
  id obj= [notif object];
  NSString *str= [obj stringValue];
  
  if (obj == colNameText)
  {
    MGRTValue column(_tableData->getColumn([columnList selectedRow]));
    column.set("name", [str UTF8String]);
    
    [self columnListChanged];
    [columnList reloadData];
  }
  else if (obj == colTypeCombo)
  {
    MGRTValue column(_tableData->getColumn([columnList selectedRow]));
    _tableData->setColumnType(column, [str UTF8String]);
    
    [columnList reloadData];
  }
  else if (obj == colDefaultText)
  {
    MGRTValue column(_tableData->getColumn([columnList selectedRow]));
    column.set("defaultValue", [str UTF8String]);
    
    [columnList reloadData];
  }
  else if (obj == indNameText)
  {
    MGRTValue index(_tableData->getIndex([indexList selectedRow]));
    index.set("name", [str UTF8String]);
    
    [indexList reloadData];
  }
  else if (obj == fkNameText)
  {
    MGRTValue fk(_tableData->getFK([fkList selectedRow]));
    fk.set("name", [str UTF8String]);
    
    [fkList reloadData];
  }  
  else
    return;
}


- (void)showColumn:(MGRTValue*)column
{  
  if (column)
  {
    char *tmp;
    [colNameText setStringValue:NSStr((const char*)(*column)["name"])];
    
    [colTypeCombo setStringValue:NSStr(tmp= _tableData->formatColumnType(*column))]; g_free(tmp);
    [colDefaultNull setState:(int)(*column)["defaultValueIsNull"] != 0 ? NSOnState : NSOffState];

    if ((int)(*column)["defaultValueIsNull"] != 0)
    {
      [colDefaultText setStringValue:@""];
      [colDefaultText setEnabled:NO];
    }
    else
    {
      [colDefaultText setStringValue:NSStr((const char*)(*column)["defaultValue"])];
      [colDefaultText setEnabled:YES];
    }
    [colFlagsList reloadData];

    [colCharsetPop selectItemAtIndex:[colCharsetPop indexOfItemWithRepresentedObject:NSStr(_tableData->getColumnCharset(*column))]];
    [self setupCollationsForCharset:_tableData->getColumnCharset(*column) popup:colCollationPop];
    [colCollationPop selectItemWithTitle:NSStr(_tableData->getColumnCollation(*column))];
    
    [colCommentText setString:NSStr((*column)["comment"].asString())];
    
    [colAIncCheck setState:(int)(*column)["autoIncrement"]?NSOnState:NSOffState];
    [colNNullCheck setState:(int)(*column)["isNullable"]==0?NSOnState:NSOffState];
    [colPKCheck setState:_tableData->columnIsPK(*column)?NSOnState:NSOffState];
    
    [colNNullCheck setEnabled:_tableData->columnIsPK(*column)?NO:YES];
    [colAIncCheck setEnabled:_tableData->columnIsNumeric(*column)?YES:NO];
  }
  else
  {
    [colNameText setStringValue:@""];
    
    [colTypeCombo setStringValue:@""];
    [colDefaultText setStringValue:@""];
    [colDefaultNull setState:NSOffState];
    [colFlagsList reloadData];
    
    [colCharsetPop selectItemWithTitle:@""];
    [colCollationPop selectItemWithTitle:@""];
    
    [colCommentText setString:@""];
    
    [colAIncCheck setState:NSOffState];
    [colNNullCheck setState:NSOffState];
    [colPKCheck setState:NSOffState];    
  }
}


- (void)showIndex:(MGRTValue*)index
{
  if (index)
  {
    [indNameText setStringValue:NSStr((const char*)(*index)["name"])];
    [indKindPop selectItemWithTitle:NSStr((const char*)(*index)["indexKind"])];
    [indTypePop selectItemWithTitle:NSStr((const char*)(*index)["indexType"])];
  }
  else
  {
    [indNameText setStringValue:@""];
    [indKindPop selectItemAtIndex:0];
    [indTypePop selectItemAtIndex:0];    
  }  
  [indColumnList reloadData];
}


- (void)showFK:(MGRTValue*)fk
{
  if (fk)
  {
    [fkNameText setStringValue:NSStr((const char*)(*fk)["name"])];
    [fkOnDeletePop selectItemWithTitle:NSStr((const char*)(*fk)["deleteRule"])];
    [fkOnUpdatePop selectItemWithTitle:NSStr((const char*)(*fk)["updateRule"])];
    [fkRefTablePop selectItemWithTitle:NSStr((const char*)(*fk)["referedTableName"])];
  }
  else
  {
    [fkNameText setStringValue:@""];
    [fkOnDeletePop selectItemAtIndex:0];
    [fkOnUpdatePop selectItemAtIndex:0];
    if ([fkRefTablePop numberOfItems]>0)
      [fkRefTablePop selectItemAtIndex:0];
  }  
  [fkColumnList reloadData];
}


- (int)numberOfRowsInTableView:(NSTableView *)aTableView
{
  if (!_tableData)
    return 0;
  
  if (aTableView == columnList)
    return _tableData->columnCount();
  else if (aTableView == colFlagsList)
  {
    int r= [columnList selectedRow];
    if (r < 0 || _tableData->columnCount()==0)
      return 0;
    return _tableData->getColumnFlags(_tableData->getColumn(r)).count();
  }
  else if (aTableView == indexList)
  {
    return _tableData->indexCount();
  }
  else if (aTableView == indColumnList)
  {
    int index= [indexList selectedRow];
    if (index < 0)
      return 0;
    return _tableData->getIndex(index)["columns"].count();
  }
  else if (aTableView == fkList)
  {
    return _tableData->fkCount();
  }
  else if (aTableView == fkColumnList)
  {
    int index= [fkList selectedRow];
    if (index < 0)
      return 0;
    return _tableData->getFK(index)["columns"].count(); 
  }
  return 0;
}


- (void)tableView:(NSTableView *)aTableView 
  willDisplayCell:(id)aCell
   forTableColumn:(NSTableColumn *)aTableColumn 
              row:(int)rowIndex
{
  if (aTableView == columnList)
  {
    MGRTValue column(_tableData->getColumn(rowIndex));
    if ([[aTableColumn identifier] isEqualTo:@"name"])
    {
      NSImage *img= _columnIcon;
      
      if (_tableData->columnIsPK(column))
        img= _pkIcon;
      else
      {
        switch (_tableData->columnTypeGroup(column))
        {
          case RDG_NUMERIC:
            img= _numericIcon;
            break;
          case RDG_DATETIME:
            img= _dateIcon;
            break;
          case RDG_STRING:
            img= _stringIcon;
            break;
          case RDG_TEXT:
            img= _stringIcon;
            break;
          case RDG_BLOB:
            img= _blobIcon;
            break;
          case RDG_GEO:
            img= _spatialIcon;
            break;
          case RDG_USER:
            img= _userdefIcon;
            break;
          case RDG_STRUCTURED:
            img= _userdefIcon;
            break;
        }
      }      
      [aCell setImage:img];
    }
    else if ([[aTableColumn identifier] isEqualTo:@"default"])
    {
      if ((int)column["defaultValueIsNull"])
        [aCell setImage:_nullIcon];
      else
        [aCell setImage:nil];
    }
  }
  else if (aTableView == colFlagsList)
  {
    int r= [columnList selectedRow];
    MGRTValue column(_tableData->getColumn(r));
    
    [aCell setTitle:NSStr(_tableData->getColumnFlags(column)[rowIndex].asString())];
  }
}


- (id)tableView:(NSTableView *)aTableView 
objectValueForTableColumn:(NSTableColumn *)aTableColumn 
            row:(int)rowIndex
{
  if (aTableView == columnList)
  {
    MGRTValue column(_tableData->getColumn(rowIndex));
    
    if ([[aTableColumn identifier] isEqualTo:@"name"])
      return NSStr((const char*)column["name"]);
    else if ([[aTableColumn identifier] isEqualTo:@"type"])
      return NSStr(_tableData->formatColumnType(column));
    else if ([[aTableColumn identifier] isEqualTo:@"nnull"])
    {
      if (_tableData->columnIsPK(column))
        return [NSNumber numberWithInt:NSOnState];
      else
        return [NSNumber numberWithInt:(int)column["isNullable"]==0?NSOnState:NSOffState];
    }
    else if ([[aTableColumn identifier] isEqualTo:@"autoinc"])
    {
      if (_tableData->columnIsNumeric(column))
        return [NSNumber numberWithInt:(int)column["autoIncrement"]?NSOnState:NSOffState];
      else
        return [NSNumber numberWithInt:NSOffState];
    }
    else if ([[aTableColumn identifier] isEqualTo:@"flags"])
    {
      int i, j;
      NSMutableString *ms= [[[NSMutableString alloc] init] autorelease];
      MGRTValue types(_tableData->getColumnFlags(column));
      j= 0;
      for (i= 0; i < types.count(); i++)
      {
        if (_tableData->columnFlagState(column, types[i].asString()))
        {
          if (j++ > 0)
            [ms appendFormat:@",%s",types[i].asString()];
          else
            [ms appendFormat:@"%s",types[i].asString()];
        }
      }
      return ms;
    }
    else if ([[aTableColumn identifier] isEqualTo:@"default"])
    {
      if ((int)column["defaultValueIsNull"])
        return @"";
      else
        return NSStr((const char*)column["defaultValue"]);
    }
    else if ([[aTableColumn identifier] isEqualTo:@"comments"])
      return NSStr((const char*)column["comment"]);
  }
  else if (aTableView == colFlagsList)
  {
    int r= [columnList selectedRow];
    int state;
    const char *flag;
    if (r < 0) return @"";
    
    MGRTValue column(_tableData->getColumn(r));
    
    flag= _tableData->getColumnFlags(column)[rowIndex];
    
    if (_tableData->columnFlagState(column, flag))
      state= NSOnState;
    else
      state= NSOffState;
    return [NSNumber numberWithInt:state];
  }
  else if (aTableView == indexList)
  {
    return NSStr((const char*)_tableData->getIndex(rowIndex)["name"]);
  }
  else if (aTableView == indColumnList)
  {
    int i= [indexList selectedRow];        
    if (i < 0)
      return nil;
    MGRTValue index(_tableData->getIndex(i));
    
    if ([[aTableColumn identifier] isEqualTo:@"column"])
    {
      int i= [[aTableColumn dataCell] indexOfItemWithTitle:NSStr((const char*)index["columns"][rowIndex]["name"])];
      if (i == NSNotFound)
        return nil;
      return [NSNumber numberWithInt:i];
    }
//    else
//      return NSStr((const char*)index["columns"][rowIndex]["length"]);
  }
  else if (aTableView == fkList)
  {
    return NSStr((const char*)_tableData->getFK(rowIndex)["name"]);
  }
  else if (aTableView == fkColumnList)
  {
    int index= [fkList selectedRow];
    if (index < 0)
      return nil;
    MGRTValue fk(_tableData->getFK(index));
    if ([[aTableColumn identifier] isEqualTo:@"column"])
    {
      int i= [[aTableColumn dataCell] indexOfItemWithTitle:NSStr(_tableData->getFKColumn(fk, rowIndex))];
      if (i < 0)
        return nil;
      return [NSNumber numberWithInt:i];
    }
    else
    {
      int i= [[aTableColumn dataCell] indexOfItemWithTitle:NSStr(_tableData->getFKRefColumn(fk, rowIndex))];
      if (i < 0)
        return nil;
      return [NSNumber numberWithInt:i];
    }
  }
  return @"";
}


- (void)tableView:(NSTableView *)aTableView 
   setObjectValue:(id)anObject
   forTableColumn:(NSTableColumn *)aTableColumn 
              row:(int)rowIndex
{
  // Column
  if (aTableView == columnList)
  {
    MGRTValue column(_tableData->getColumn(rowIndex));
    
    if ([[aTableColumn identifier] isEqualTo:@"name"])
    {
      column.set("name", [anObject UTF8String]);
      [self columnListChanged];
    }
    else if ([[aTableColumn identifier] isEqualTo:@"type"])
    {
      _tableData->setColumnType(column, [anObject UTF8String]);
    }
    else if ([[aTableColumn identifier] isEqualTo:@"nnull"])
    {
      column.set("isNullable", [anObject intValue] != NSOnState ? 1 : 0);
    }
    else if ([[aTableColumn identifier] isEqualTo:@"autoinc"])
    {
      if (_tableData->columnIsNumeric(column))
        column.set("autoIncrement", [anObject intValue] == NSOnState ? 1 : 0);
    }  
    else if ([[aTableColumn identifier] isEqualTo:@"flags"])
    {
      MGRTValue types(_tableData->getColumnFlags(column));
      int i, t;
      char **tokens= g_strsplit([anObject UTF8String], ",", 0);
      
      for (i= 0; i < types.count(); i++)
      {
        bool on= false;
        
        for (t= 0; tokens[t]; t++)
        {
          if (strcasecmp(types[i].asString(), tokens[t])==0)
          {
            on= true;
            break;
          }
        }
        _tableData->setColumnFlagState(column, types[i].asString(), on);
      }      
    }
    else if ([[aTableColumn identifier] isEqualTo:@"default"])
    {
      column.set("defaultValue", [anObject UTF8String]);
      if ([colDefaultNull state] == NSOnState)
        [colDefaultNull performClick:nil];
      column.set("defaultValueIsNull", 0);
    }
    else if ([[aTableColumn identifier] isEqualTo:@"comments"])
    {
      column.set("comment", [anObject UTF8String]);
    }

    [self showColumn:&column];
  }
  // Column Flags
  else if (aTableView == colFlagsList)
  {
    int r= [columnList selectedRow];
    MGRTValue column(_tableData->getColumn(r));
    const char *flag= _tableData->getColumnFlags(column)[rowIndex];
    
    _tableData->setColumnFlagState(column, flag, [anObject intValue]==NSOnState);
    [columnList reloadData];
  }
  // FK Columns
  else if (aTableView == fkColumnList)
  {
    int index= [fkList selectedRow];
    int cindex= [fkColumnList selectedRow];
    MGRTValue fk(_tableData->getFK(index));
    int i= [anObject intValue];
    NSString *name;
    
    if (i >= 0)
    {
      name= [[aTableColumn dataCell] itemTitleAtIndex:i];
      
      if ([[aTableColumn identifier] isEqualTo:@"column"])
        _tableData->setFKColumn(fk, cindex, [name UTF8String]);
      else
        _tableData->setFKRefColumn(fk, cindex, [name UTF8String]);
      [fkColumnList reloadData];
    }
  }  
  // Index Columns
  else if (aTableView == indColumnList)
  {
    int i= [indexList selectedRow];
    int cindex= [indColumnList selectedRow];
    MGRTValue index(_tableData->getIndex(i));

    if ([[aTableColumn identifier] isEqualTo:@"column"])
    {
      int i= [anObject intValue];
      NSString *name= [[aTableColumn dataCell] itemTitleAtIndex:i];
      
      _tableData->setIndexColumn(index, cindex, [name UTF8String]);
    }
    else
    {
      /* length
      g_free(cdata->len);
      if ([anObject intValue]==0)
        cdata->len= g_strdup("");
      else
        cdata->len= g_strdup([anObject UTF8String]);
       */
    }
    [indColumnList reloadData];
  }
}

- (BOOL)tableView:(NSTableView *)tableView
        writeRows:(NSArray *)rows
     toPasteboard:(NSPasteboard *)pboard
{
  if (tableView == columnList)
  {
    _draggedRow= [[rows objectAtIndex:0] intValue];
    
    [pboard declareTypes:[NSArray arrayWithObject:MGRTTableRowPboardType] owner:nil];
    [pboard setData:[NSData data] forType:MGRTTableRowPboardType];
    
    return YES;
  }
  else
    return NO;
}

- (NSDragOperation)tableView:(NSTableView *)tableView
                validateDrop:(id <NSDraggingInfo>)info
                 proposedRow:(int)row
       proposedDropOperation:(NSTableViewDropOperation)operation
{
  NSDragOperation op;
  if (tableView == columnList)
  {
    if ([info draggingSource] == tableView && row != _draggedRow && row >= 0)
      op= NSDragOperationPrivate;
    else
      op= NSDragOperationNone;
  }
  else if (tableView == fkColumnList && [fkColumnList isEnabled] && [fkList selectedRow]>=0)
    op= NSDragOperationPrivate;
  else if (tableView == indColumnList && [indColumnList isEnabled] && [indexList selectedRow]>=0)
    op= NSDragOperationPrivate;
  else 
    op= NSDragOperationNone;
  return op;
}

- (BOOL)tableView:(NSTableView *)tableView 
       acceptDrop:(id <NSDraggingInfo>)info
              row:(int)row 
    dropOperation:(NSTableViewDropOperation)operation
{
  if (tableView == columnList)
  {
    if (row >= _tableData->columnCount())
      row= _tableData->columnCount()-1;
    
    _tableData->moveColumn(_draggedRow, row);
    
    [tableView reloadData];
    return YES;
  }
  else if (tableView == fkColumnList)
  {
    int i= [fkList selectedRow];
    MGRTValue fk(_tableData->getFK(i));
    int findex;
    
    [self addFKColumn:nil];
    findex= [fkColumnList selectedRow];
    
    _tableData->setFKColumn(fk, findex, (const char*)_tableData->getColumn(_draggedRow)["name"]);
    
    [fkColumnList reloadData];
    return YES;    
  }
  else if (tableView == indColumnList)
  {
    int cindex;
    int i= [indexList selectedRow];
    MGRTValue index(_tableData->getIndex(i));
    
    [self addIndexColumn:nil];
    cindex= [indColumnList selectedRow];
    
    index["columns"][cindex].set("name", (const char*)_tableData->getColumn(_draggedRow)["name"]);
    
    [fkColumnList reloadData];
    return YES;
  }
  else
    return NO;
}

- (void)tableViewSelectionDidChange:(NSNotification *)aNotification
{
  NSTableView *table= [aNotification object];
  int row= [table selectedRow];
  BOOL enabled= NO;

  if (table == columnList)
  {  
    if (row < 0)
    {
      [colGroup setEnabledRecursive:NO];
      [self showColumn:0];
    }
    else
    {
      MGRTValue column(_tableData->getColumn(row));
      [colGroup setEnabledRecursive:YES];
      [self showColumn:&column];
    }
  }
  else if (table == indexList)
  {
    if (row < 0)
    {
      enabled= NO;
      [self showIndex:0];
    }
    else
    {
      MGRTValue index(_tableData->getIndex(row));
      enabled= YES;
      [self showIndex:&index];
    }
    [indBox setEnabledRecursive:enabled];
  }
  else if (table == fkList)
  {
    if (row < 0)
    {
      enabled= NO;
      [self showFK:0];
    }
    else
    {
      MGRTValue fk(_tableData->getFK(row));
      enabled= YES;
      [self showFK:&fk];
    }
    [fkBox setEnabledRecursive:enabled];
  }  
}


- (void)setupCollationsForCharset:(const char*)charset
                            popup:(id)popup
{
  int i;

  [popup removeAllItems];
  
  if (charset)
  {
    std::vector<const char*>collations= _tableData->getCharsetCollations((*_catalogs)["characterSets"], charset);
    
    for (i= 0; i < collations.size(); i++)
    {
      [popup addItemWithTitle:NSStr(collations[i])];
    }
  }
}

- (void)showObject
{
  MGRTValue table(_tableData->value());
  
  // Column Stuff
  [tableNameText setStringValue:NSStr(table["name"].asString())];
  [commentText setStringValue:NSStr(table["comment"].asString())];
  
  [self showColumn:0];
  [columnList deselectAll:nil];
  [colGroup setEnabledRecursive:NO];  
  [columnList reloadData];
  [self columnListChanged];
  
  // Index Stuff
  [self showIndex:0];
  [indexList deselectAll:nil];
  [indBox setEnabledRecursive:NO];
  [indexList reloadData];
  
  // FK Stuff
  [self showFK:0];
  [fkList deselectAll:nil];
  [fkBox setEnabledRecursive:NO];
  [fkList reloadData];
  
  // Engine Stuff
  const char *engine= table["tableEngine"];
  for (int i= 0; i < [engineMatrix numberOfRows]; i++)
  {
    if (strcmp2([[[engineMatrix cellAtRow:i column:0] title] UTF8String],
                engine)==0)
    {
      [[engineMatrix cellAtRow:i column:0] performClick:nil];
      break;
    }
  }
  [self refreshForTableType];
  
  // Charset Stuff
  [tblCharsetPop selectItemWithTitle:NSStr((const char*)table["defaultCharacterSetName"])];
  [self setupCollationsForCharset:table["defaultCharacterSetName"] popup:tblCollationPop];
  [tblCollationPop selectItemWithTitle:NSStr((const char*)table["defaultCollationName"])];
  
  // Misc. Options
  [[miscOpView viewWithTag:1] selectItemWithTitle:NSStr(table.get("packKeys", ""))];
  [[miscOpView viewWithTag:2] setStringValue:NSStr(table.get("password", ""))];
  [[miscOpView viewWithTag:3] setStringValue:NSStr(table.get("nextAutoInc", ""))];
  [[miscOpView viewWithTag:4] setState:table.get("delayKeyWrite",0)?NSOnState:NSOffState];

  // Row Options
  [[rowOpView viewWithTag:1] selectItemWithTitle:NSStr(table.get("rowFormat", ""))];
  [[rowOpView viewWithTag:2] setState:table.get("checksum",0)?NSOnState:NSOffState];
  [[rowOpView viewWithTag:3] setStringValue:NSStr(table.get("avgRowLength",""))];
  [[rowOpView viewWithTag:4] setStringValue:NSStr(table.get("minRows",""))];
  [[rowOpView viewWithTag:5] setStringValue:NSStr(table.get("maxRows",""))];

  // Storage Options
  [[stoOpView viewWithTag:1] setStringValue:NSStr(table.get("tableDataDir",""))];
  [[stoOpView viewWithTag:3] setStringValue:NSStr(table.get("tableIndexDir",""))];

  // Merge Options
  [[mergeOpView viewWithTag:1] setStringValue:NSStr(table.get("mergeUnion",""))];
  [[mergeOpView viewWithTag:2] selectItemWithTitle:NSStr(table.get("mergeInsert",""))];

  // Table RAID stuff
  [[raidOpView viewWithTag:1] selectItemWithTitle:NSStr(table.get("raidType",""))];
  [[raidOpView viewWithTag:2] setStringValue:NSStr(table.get("raidChunks",""))];
  [[raidOpView viewWithTag:3] setStringValue:NSStr(table.get("raidChunkSize",""))];
}


- (BOOL)commitChanges
{
  MGRTValue table(_tableData->value());
  
#define ASSIGNSTR(v,s) table.set(v, [s UTF8String])
  
  ASSIGNSTR("name", [tableNameText stringValue]);
  ASSIGNSTR("comment", [commentText stringValue]);
  
  // Misc. Options
  ASSIGNSTR("packKeys", [[miscOpView viewWithTag:1] titleOfSelectedItem]);
  ASSIGNSTR("password", [[miscOpView viewWithTag:2] stringValue]);
  ASSIGNSTR("nextAutoInc", [[miscOpView viewWithTag:3] stringValue]);
  table.set("delayKeyWrite", [[miscOpView viewWithTag:4] state] == NSOnState);
  
  // Row Options
  ASSIGNSTR("rowFormat", [[rowOpView viewWithTag:1] titleOfSelectedItem]);
  table.set("checksum", [[rowOpView viewWithTag:2] state] == NSOnState);
  ASSIGNSTR("avgRowLength", [[rowOpView viewWithTag:3] stringValue]);
  ASSIGNSTR("minRows", [[rowOpView viewWithTag:4] stringValue]);
  ASSIGNSTR("maxRows", [[rowOpView viewWithTag:5] stringValue]);
  
  // Storage Options
  ASSIGNSTR("tableDataDir", [[stoOpView viewWithTag:1] stringValue]);
  ASSIGNSTR("tableIndexDir", [[stoOpView viewWithTag:3] stringValue]);
  
  // Merge Options
  ASSIGNSTR("mergeUnion", [[mergeOpView viewWithTag:1] stringValue]);
  ASSIGNSTR("mergeInsert", [[mergeOpView viewWithTag:2] titleOfSelectedItem]);
  
  // Table RAID stuff
  ASSIGNSTR("raidType", [[raidOpView viewWithTag:1] titleOfSelectedItem]);
  ASSIGNSTR("raidChunks", [[raidOpView viewWithTag:2] stringValue]);
  ASSIGNSTR("raidChunkSize", [[raidOpView viewWithTag:3] stringValue]);

  return YES;
}


- (void)refreshForTableType
{
  BOOL fkEnabled= NO;

  fkEnabled= strcasecmp(_tableData->value()["tableEngine"].asString(),
                        "InnoDB")==0;
  
  [fkContainer setEnabledRecursive:fkEnabled];
  if (fkEnabled)
    [fkBox setEnabledRecursive:NO];
}

@end

//===

@implementation MGRTTableEditor

- (IBAction)applyChanges:(id)sender
{
  if ([self commitChanges])
  {
    _tableData->commit();
  }
}

- (IBAction)close:(id)sender
{
  _tableData->revert();

  [self close];
}

- (IBAction)discardChanges:(id)sender
{
  _tableData->revert();

  [self showObject];
}

- (IBAction)handleValueChange:(id)sender
{
  if (sender == tblCharsetPop)
  {
    const char *charset= [[[sender selectedItem] representedObject] UTF8String];
    
    [self setupCollationsForCharset:charset popup:tblCollationPop];
    
    _tableData->value().set("defaultCharacterSetName", charset);
  }  
  else if (sender == tblCollationPop)
  {
    const char *collation= [[sender titleOfSelectedItem] UTF8String];
    
    _tableData->value().set("defaultCollationName", collation);
  }
  else if (sender == engineMatrix)
  {
    _tableData->value().set("tableEngine", [[[engineMatrix selectedCell] title] UTF8String]);
    [self refreshForTableType];
  }
}

- (IBAction)handleColumnValueChange:(id)sender
{
  int column= [columnList selectedRow];

  if (sender == colCharsetPop)
  {
    const char *charset= [[[sender selectedItem] representedObject] UTF8String];
    
    [self setupCollationsForCharset:charset popup:colCollationPop];
    _tableData->getColumn(column).set("characterSetName", charset);
  }  
  else if (sender == colCollationPop)
  {
    const char *collation= [[sender titleOfSelectedItem] UTF8String];
    
    _tableData->getColumn(column).set("collationName", collation);
  }
  else if (sender == colDefaultNull)
  {
    _tableData->getColumn(column).set("defaultValueIsNull", [sender state] == NSOnState);
    if ([sender state] == NSOnState)
    {
      [colDefaultText setEnabled:NO];
    }
    else
    {
      _tableData->getColumn(column).set("defaultValue", "");
      [colDefaultText setEnabled:YES];
    }
    [columnList reloadData];
  }
  else if ([sender isMemberOfClass:[NSMatrix class]])
  {
    MGRTValue gcolumn(_tableData->getColumn(column));
    
    _tableData->setColumnPK(gcolumn, [colPKCheck state] == NSOnState);
    
    if (!_tableData->columnIsNumeric(gcolumn))
      gcolumn.set("autoIncrement", 0);
    else
      gcolumn.set("autoIncrement", [colAIncCheck state] == NSOnState ? 1 : 0);

    if (_tableData->columnIsPK(gcolumn))
    {
      gcolumn.set("isNullable", 0);
      [colNNullCheck setState:NSOnState];
      [colNNullCheck setEnabled:NO];
    }
    else
    {
      gcolumn.set("isNullable", [colNNullCheck state] != NSOnState ? 1 : 0);
      [colNNullCheck setEnabled:YES];
    }
    
    [columnList reloadData];
  }
}

- (IBAction)handleIndexValueChange:(id)sender
{
  int i= [indexList selectedRow];
  MGRTValue index(_tableData->getIndex(i));
  
  if (sender == indKindPop)
  {
    index.set("indexKind", [[sender titleOfSelectedItem] UTF8String]);
  }
  else if (sender == indTypePop)
  {
    index.set("indexType", [[sender titleOfSelectedItem] UTF8String]);
  }
  else
    NSAssert(0, @"Invalid sender!");
}


- (IBAction)handleFKValueChange:(id)sender
{
  int f= [fkList selectedRow];
  MGRTValue fk(_tableData->getFK(f));
  
  if (sender == fkOnUpdatePop)
  {
    fk.set("updateRule", [[sender titleOfSelectedItem] UTF8String]);
  }
  else if (sender == fkOnDeletePop)
  {
    fk.set("deleteRule", [[sender titleOfSelectedItem] UTF8String]);
  }
  else if (sender == fkRefTablePop)
  {
    NSString *table= [sender titleOfSelectedItem];
    
    if (strcmp2([table UTF8String], (const char*)fk["referedTableName"])!=0)
    {
      int i;
      id cell;
      MGRTValue refTable;
      
      fk.set("referedTableName", [table UTF8String]);

      _tableData->resetFKColumns(fk);
      
      // setup the menu in the column dataCell
      cell= [[fkColumnList tableColumnWithIdentifier:@"fcolumn"] dataCell];
      
      //XXX
      // fetch the column names for the table
      refTable= (*_catalogs)["schemata"].listItemNamed("test")["tables"].listItemNamed([table UTF8String]);
      
      if (refTable.isValid())
      {
        [cell removeAllItems];
        for (i= 0; i < refTable["columns"].count(); i++)
          [cell addItemWithTitle:NSStr((const char*)refTable["columns"][i]["name"])];
      }
      else
        NSLog(@"Could not find table %@ in catalogs list", table);
    }
  }
}


- (IBAction)chooseFile:(id)sender
{
  NSOpenPanel *panel= [NSOpenPanel openPanel];
  
  [panel setCanChooseFiles:NO];
  [panel setCanChooseDirectories:YES];
  [panel setTitle:@"Choose Directory"];
  if ([panel runModalForDirectory:nil file:nil types:nil]==NSOKButton)
  {
    NSString *path= [panel directory];
    
    [[[sender superview] viewWithTag:([sender tag] - 1)] setStringValue:path];
  }
}


- (id)init
{
  self= [super initWithWindowNibName: @"TableEditor"];
  if (self)
  {
    [self loadWindow];
    
    [[NSNotificationCenter defaultCenter] addObserver:self
                                             selector:@selector(handleTextEdited:)
                                                 name:NSControlTextDidChangeNotification
                                               object:nil];
    [[self window] setDelegate:self];
  }
  return self;
}


- (void)setGRT:(MGRT*)grt catalogs:(MGRTValue)catalogs
{
  _grt= [grt retain];
  
  _catalogs= new MGRTValue(catalogs);
}


- (void)dealloc
{
  delete _catalogs;
  
  [_grt release];
  
  [[NSNotificationCenter defaultCenter] removeObserver:self];
  
  [_pkIcon release];
  [_columnIcon release];
  [_blobIcon release];
  [_dateIcon release];
  [_numericIcon release];
  [_spatialIcon release];
  [_stringIcon release];
  [_userdefIcon release];
  
  [super dealloc];
}


- (void)setupButtonsForTableView:(NSTableView*)tview
                       addAction:(SEL)addSel
                    deleteAction:(SEL)delSel
{
  NSButton *button;
  MAccessoryScrollView *sv= (MAccessoryScrollView*)[tview enclosingScrollView];
  float size;

//  size= [NSScroller scrollerWidthForControlSize:NSMiniControlSize];
  size= 16;

  button= [[[NSButton alloc] initWithFrame:NSMakeRect(0,0,size,size)] autorelease];
  [button setButtonType: NSMomentaryChangeButton];
  [button setBordered:NO];
  [[button cell] setImagePosition:NSImageOnly];
  [button setImage: MXGetImageFromBundle([NSBundle bundleForClass:[self class]], @"mini_add.png")];
  [button setAlternateImage: MXGetImageFromBundle([NSBundle bundleForClass:[self class]], @"mini_add_pressed.png")];

  [button setTarget:self];
  [button setAction:addSel];
  [sv addSubview:button];
  [sv addHorizontalAccessoryView:button];
  
  button= [[[NSButton alloc] initWithFrame:NSMakeRect(0,0,size,size)] autorelease];
  [button setButtonType: NSMomentaryChangeButton];
  [button setBordered:NO];
  [[button cell] setImagePosition:NSImageOnly];
  [button setImage: MXGetImageFromBundle([NSBundle bundleForClass:[self class]], @"mini_del.png")];
  [button setAlternateImage: MXGetImageFromBundle([NSBundle bundleForClass:[self class]], @"mini_del_pressed.png")];
  [button setTarget:self];
  [button setAction:delSel];
  [sv addSubview:button];
  [sv addHorizontalAccessoryView:button];
}


- (void)awakeFromNib
{
  NSBundle *bundle= [NSBundle bundleForClass:[self class]];
  
  [self setupButtonsForTableView:columnList 
                       addAction:@selector(addColumn:)
                    deleteAction:@selector(deleteColumn:)];
  [self setupButtonsForTableView:fkList
                       addAction:@selector(addFK:)
                    deleteAction:@selector(deleteFK:)];
  [self setupButtonsForTableView:fkColumnList
                       addAction:@selector(addFKColumn:)
                    deleteAction:@selector(deleteFKColumn:)];
  [self setupButtonsForTableView:indexList
                       addAction:@selector(addIndex:)
                    deleteAction:@selector(deleteIndex:)];
  [self setupButtonsForTableView:indColumnList
                       addAction:@selector(addIndexColumn:)
                    deleteAction:@selector(deleteIndexColumn:)];

  [[columnList tableColumnWithIdentifier:@"nnull"] setHeaderCell:
    [[NSTableHeaderCell alloc] initImageCell:MXGetImageFromBundle(bundle,@"editor_table_not_null.png")]];
  [[columnList tableColumnWithIdentifier:@"autoinc"] setHeaderCell:
    [[NSTableHeaderCell alloc] initImageCell:MXGetImageFromBundle(bundle,@"editor_table_auto_inc.png")]];

  _pkIcon= [MXGetImageFromBundle(bundle,@"column_pk.png") retain];
  _columnIcon= [MXGetImageFromBundle(bundle,@"column.png") retain];
  _blobIcon= [MXGetImageFromBundle(bundle,@"datatype_blob.png") retain];
  _dateIcon= [MXGetImageFromBundle(bundle,@"datatype_datetime.png") retain];
  _numericIcon= [MXGetImageFromBundle(bundle,@"datatype_numeric.png") retain];
  _spatialIcon= [MXGetImageFromBundle(bundle,@"datatype_spatial.png") retain];
  _stringIcon= [MXGetImageFromBundle(bundle,@"datatype_string.png") retain];
  _userdefIcon= [MXGetImageFromBundle(bundle,@"datatype_userdefined.png") retain];
  
  _nullIcon= [MXGetImageFromBundle(bundle,@"field_overlay_null.png") retain];
  
  [columnList registerForDraggedTypes:[NSArray arrayWithObject:MGRTTableRowPboardType]];
  [fkColumnList registerForDraggedTypes:[NSArray arrayWithObject:MGRTTableRowPboardType]];
  [indColumnList registerForDraggedTypes:[NSArray arrayWithObject:MGRTTableRowPboardType]];
  
  [colDefaultNull setToolTip:@"Make default value NULL."];  
}


- (void)editObject:(MYX_GRT_VALUE*)value
{
  _tableData= new MGRTTable([_grt grt], value);
  
  // set up the Schema selection popup button
  [schemaPop removeAllItems];

  MGRTValue schema(_tableData->ownerSchema());
  MGRTValue schemata((*_catalogs)["schemata"]);
  int i, c;
  
  c= schemata.count();
  for (i= 0; i < c; i++)
  {
    [schemaPop addItemWithTitle:NSStr(schemata[i]["name"].asString())];
    
    if (strcmp(schemata[i]["_id"].asString(), schema["_id"].asString())==0)
      [schemaPop selectItem:[schemaPop lastItem]];
  }

  // we dont allow it to be changed by now
  [schemaPop setEnabled:NO];

  // fill the table popup for FK setup 
  [fkRefTablePop removeAllItems];
  MGRTValue tables(schema["tables"]);
  c= tables.count();
  for (i= 0; i < c; i++)
    [fkRefTablePop addItemWithTitle:NSStr(tables[i]["name"].asString())];

  // column types
  MGRTValue simpleTypes(MGRTValue::fromGlobal([_grt grt], "/rdbmsMgmt/rdbms/Mysql/simpleDatatypes"));
  
  [colTypeCombo removeAllItems];
  c= simpleTypes.count();
  g_message("types%i", c);
  for (i= 0; i < c; i++)
  {
    NSString *tmp;
    if (simpleTypes[i]["numericPrecision"].asInt() > 0)
    {
      tmp= [NSString stringWithFormat:@"%s(11)", simpleTypes[i]["name"].asString()];
    }
    else if (simpleTypes[i]["characterMaximumLength"].asInt() > 0)
    {
      tmp= [NSString stringWithFormat:@"%s(40)", simpleTypes[i]["name"].asString()];
    }
    else
      tmp= NSStr(simpleTypes[i]["name"].asString());
    [colTypeCombo addItemWithObjectValue:tmp];
  }

  // charset and collations
  [colCharsetPop removeAllItems];
  [tblCharsetPop removeAllItems];
  [colCharsetPop addItemWithTitle:@""];
  [[colCharsetPop lastItem] setRepresentedObject:@""];
  [tblCharsetPop addItemWithTitle:@""];
  
  MGRTValue charsets((*_catalogs)["characterSets"]);
  c= charsets.count();
  for (i= 0; i < c; i++)
  {
    NSString *name= NSStr(charsets[i]["name"].asString());
    NSString *descr= NSStr(charsets[i]["description"].asString());
    
    [colCharsetPop addItemWithTitle:descr];
    [[colCharsetPop lastItem] setRepresentedObject:name];
    [tblCharsetPop addItemWithTitle:descr];
    [[tblCharsetPop lastItem] setRepresentedObject:name];
  }
  
  [self showObject];
}


- (IBAction)confirmSave:(id)sender
{
}

- (IBAction)confirmCancel:(id)sender
{
}

- (IBAction)confirmOK:(id)sender
{
}

- (void)setReleaseOnClose:(BOOL)flag
{
  _releaseOnClose= flag;
}

- (void)windowWillClose:(NSNotification*)notification
{
  if (_releaseOnClose)
    [self autorelease];
}

@end

extern "C" {
  MYX_GRT_VALUE *grt_openTableEditor(MYX_GRT_VALUE *arg, void *udata)
  {
    MGRTTableEditor *editor= [[MGRTTableEditor alloc] init];
    
    NSLog(@"OPEN TABLE EDI");
    
    return 0;
  }
};
